# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/22 下午8:06
Written by Yuwei Jin(yuwei_jin@163.com)
"""

from yacs.config import CfgNode as CN


# configs for HRNet48
HRNET_48 = CN()
HRNET_48.FINAL_CONV_KERNEL = 1

HRNET_48.STAGE1 = CN()
HRNET_48.STAGE1.NUM_MODULES = 1
HRNET_48.STAGE1.NUM_BRANCHES = 1
HRNET_48.STAGE1.NUM_BLOCKS = [4]
HRNET_48.STAGE1.NUM_CHANNELS = [64]
HRNET_48.STAGE1.BLOCK = 'BOTTLENECK'
HRNET_48.STAGE1.FUSE_METHOD = 'SUM'

HRNET_48.STAGE2 = CN()
HRNET_48.STAGE2.NUM_MODULES = 1
HRNET_48.STAGE2.NUM_BRANCHES = 2
HRNET_48.STAGE2.NUM_BLOCKS = [4, 4]
HRNET_48.STAGE2.NUM_CHANNELS = [48, 96]
HRNET_48.STAGE2.BLOCK = 'BASIC'
HRNET_48.STAGE2.FUSE_METHOD = 'SUM'

HRNET_48.STAGE3 = CN()
HRNET_48.STAGE3.NUM_MODULES = 4
HRNET_48.STAGE3.NUM_BRANCHES = 3
HRNET_48.STAGE3.NUM_BLOCKS = [4, 4, 4]
HRNET_48.STAGE3.NUM_CHANNELS = [48, 96, 192]
HRNET_48.STAGE3.BLOCK = 'BASIC'
HRNET_48.STAGE3.FUSE_METHOD = 'SUM'

HRNET_48.STAGE4 = CN()
HRNET_48.STAGE4.NUM_MODULES = 3
HRNET_48.STAGE4.NUM_BRANCHES = 4
HRNET_48.STAGE4.NUM_BLOCKS = [4, 4, 4, 4]
HRNET_48.STAGE4.NUM_CHANNELS = [48, 96, 192, 384]
HRNET_48.STAGE4.BLOCK = 'BASIC'
HRNET_48.STAGE4.FUSE_METHOD = 'SUM'


# configs for HRNet40
HRNET_40 = CN()
HRNET_40.FINAL_CONV_KERNEL = 1

HRNET_40.STAGE1 = CN()
HRNET_40.STAGE1.NUM_MODULES = 1
HRNET_40.STAGE1.NUM_BRANCHES = 1
HRNET_40.STAGE1.NUM_BLOCKS = [4]
HRNET_40.STAGE1.NUM_CHANNELS = [64]
HRNET_40.STAGE1.BLOCK = 'BOTTLENECK'
HRNET_40.STAGE1.FUSE_METHOD = 'SUM'

HRNET_40.STAGE2 = CN()
HRNET_40.STAGE2.NUM_MODULES = 1
HRNET_40.STAGE2.NUM_BRANCHES = 2
HRNET_40.STAGE2.NUM_BLOCKS = [4, 4]
HRNET_40.STAGE2.NUM_CHANNELS = [40, 80]
HRNET_40.STAGE2.BLOCK = 'BASIC'
HRNET_40.STAGE2.FUSE_METHOD = 'SUM'

HRNET_40.STAGE3 = CN()
HRNET_40.STAGE3.NUM_MODULES = 4
HRNET_40.STAGE3.NUM_BRANCHES = 3
HRNET_40.STAGE3.NUM_BLOCKS = [4, 4, 4]
HRNET_40.STAGE3.NUM_CHANNELS = [40, 80, 160]
HRNET_40.STAGE3.BLOCK = 'BASIC'
HRNET_40.STAGE3.FUSE_METHOD = 'SUM'

HRNET_40.STAGE4 = CN()
HRNET_40.STAGE4.NUM_MODULES = 3
HRNET_40.STAGE4.NUM_BRANCHES = 4
HRNET_40.STAGE4.NUM_BLOCKS = [4, 4, 4, 4]
HRNET_40.STAGE4.NUM_CHANNELS = [40, 80, 160, 320]
HRNET_40.STAGE4.BLOCK = 'BASIC'
HRNET_40.STAGE4.FUSE_METHOD = 'SUM'


# configs for HRNet32
HRNET_32 = CN()
HRNET_32.FINAL_CONV_KERNEL = 1

HRNET_32.STAGE1 = CN()
HRNET_32.STAGE1.NUM_MODULES = 1
HRNET_32.STAGE1.NUM_BRANCHES = 1
HRNET_32.STAGE1.NUM_BLOCKS = [4]
HRNET_32.STAGE1.NUM_CHANNELS = [64]
HRNET_32.STAGE1.BLOCK = 'BOTTLENECK'
HRNET_32.STAGE1.FUSE_METHOD = 'SUM'

HRNET_32.STAGE2 = CN()
HRNET_32.STAGE2.NUM_MODULES = 1
HRNET_32.STAGE2.NUM_BRANCHES = 2
HRNET_32.STAGE2.NUM_BLOCKS = [4, 4]
HRNET_32.STAGE2.NUM_CHANNELS = [32, 64]
HRNET_32.STAGE2.BLOCK = 'BASIC'
HRNET_32.STAGE2.FUSE_METHOD = 'SUM'

HRNET_32.STAGE3 = CN()
HRNET_32.STAGE3.NUM_MODULES = 4
HRNET_32.STAGE3.NUM_BRANCHES = 3
HRNET_32.STAGE3.NUM_BLOCKS = [4, 4, 4]
HRNET_32.STAGE3.NUM_CHANNELS = [32, 64, 128]
HRNET_32.STAGE3.BLOCK = 'BASIC'
HRNET_32.STAGE3.FUSE_METHOD = 'SUM'

HRNET_32.STAGE4 = CN()
HRNET_32.STAGE4.NUM_MODULES = 3
HRNET_32.STAGE4.NUM_BRANCHES = 4
HRNET_32.STAGE4.NUM_BLOCKS = [4, 4, 4, 4]
HRNET_32.STAGE4.NUM_CHANNELS = [32, 64, 128, 256]
HRNET_32.STAGE4.BLOCK = 'BASIC'
HRNET_32.STAGE4.FUSE_METHOD = 'SUM'


# configs for HRNet18
HRNET_18 = CN()
HRNET_18.FINAL_CONV_KERNEL = 1

HRNET_18.STAGE1 = CN()
HRNET_18.STAGE1.NUM_MODULES = 1
HRNET_18.STAGE1.NUM_BRANCHES = 1
HRNET_18.STAGE1.NUM_BLOCKS = [4]
HRNET_18.STAGE1.NUM_CHANNELS = [64]
HRNET_18.STAGE1.BLOCK = 'BOTTLENECK'
HRNET_18.STAGE1.FUSE_METHOD = 'SUM'

HRNET_18.STAGE2 = CN()
HRNET_18.STAGE2.NUM_MODULES = 1
HRNET_18.STAGE2.NUM_BRANCHES = 2
HRNET_18.STAGE2.NUM_BLOCKS = [4, 4]
HRNET_18.STAGE2.NUM_CHANNELS = [18, 36]
HRNET_18.STAGE2.BLOCK = 'BASIC'
HRNET_18.STAGE2.FUSE_METHOD = 'SUM'

HRNET_18.STAGE3 = CN()
HRNET_18.STAGE3.NUM_MODULES = 4
HRNET_18.STAGE3.NUM_BRANCHES = 3
HRNET_18.STAGE3.NUM_BLOCKS = [4, 4, 4]
HRNET_18.STAGE3.NUM_CHANNELS = [18, 36, 72]
HRNET_18.STAGE3.BLOCK = 'BASIC'
HRNET_18.STAGE3.FUSE_METHOD = 'SUM'

HRNET_18.STAGE4 = CN()
HRNET_18.STAGE4.NUM_MODULES = 3
HRNET_18.STAGE4.NUM_BRANCHES = 4
HRNET_18.STAGE4.NUM_BLOCKS = [4, 4, 4, 4]
HRNET_18.STAGE4.NUM_CHANNELS = [18, 36, 72, 144]
HRNET_18.STAGE4.BLOCK = 'BASIC'
HRNET_18.STAGE4.FUSE_METHOD = 'SUM'


MODEL_CONFIGS = {
    'hrnet18': HRNET_18,
    'hrnet32': HRNET_32,
    'hrnet40': HRNET_40,
    'hrnet48': HRNET_48,
}


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 18:29 
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
from torch import nn

from sv.libs.tools.init_weight import init_weights


class CrossLevelGate(nn.Module):
    def __init__(self, in_chs):
        super().__init__()
        self.gate = nn.Sequential(
            nn.Conv2d(in_chs, 1, kernel_size=1),
            nn.Sigmoid()
        )
        self.gamma = nn.Parameter(torch.ones(1))

        init_weights(self.gate)

    def forward(self, low_f, high_f):
        x = torch.cat([low_f, high_f], dim=1)
        out = self.gate(x) * self.gamma * low_f
        return out


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-23 13:24 
Written by Yuwei Jin (642281525@qq.com)
"""
import argparse
import os
import torch
import torch.nn.parallel
import torch.optim
import torch.utils.data
import torch.utils.data.distributed

from rep_vgg import get_RepVGG_func_by_name, repvgg_model_convert

parser = argparse.ArgumentParser(description='RepVGG Conversion')
parser.add_argument('load', metavar='LOAD', help='path to the weights file')
parser.add_argument('save', metavar='SAVE', help='path to the weights file')
parser.add_argument('-a', '--arch', metavar='ARCH', default='RepVGG-A0')

def convert():
    args = parser.parse_args()

    repvgg_build_func = get_RepVGG_func_by_name(args.arch)

    train_model = repvgg_build_func(deploy=False)

    if os.path.isfile(args.load):
        print("=> loading checkpoint '{}'".format(args.load))
        checkpoint = torch.load(args.load)
        if 'state_dict' in checkpoint:
            checkpoint = checkpoint['state_dict']
        elif 'model' in checkpoint:
            checkpoint = checkpoint['model']
        ckpt = {k.replace('module.', ''): v for k, v in checkpoint.items()}  # strip the names
        print(ckpt.keys())
        train_model.load_state_dict(ckpt)
    else:
        print("=> no checkpoint found at '{}'".format(args.load))

    repvgg_model_convert(train_model, save_path=args.save)


if __name__ == '__main__':
    convert()


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 17:14 
Written by Yuwei Jin (642281525@qq.com)
"""

import torch.nn as nn
from sv.libs.tools.init_weight import init_weights


class SEBlock(nn.Module):
    def __init__(self, in_chs, r=16):
        super(SEBlock, self).__init__()

        self.gap = nn.AdaptiveAvgPool2d(1)

        self.fc = nn.Sequential(
            nn.Conv2d(in_chs, in_chs // r, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_chs // r, in_chs, 1),
            nn.Sigmoid()
        )

        init_weights(self.fc)

    def forward(self, x):
        gap = self.gap(x)
        weight = self.fc(gap)
        return weight * x



# -*- encoding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021/4/11 下午3:48
Written by Yuwei Jin(yuwei_jin@163.com)
"""

import numpy as np


class Metrics(object):
    """Base metrics for pixel-wise segmentation task.
    implemented metrics including:
        1. class-wise pixel accuracy (precision)
        2. class-wise recall
        3. class-wise F1-score
        4. overall accuracy
        5. class-wise IoU
        6. mean IoU
        7. class-wise frequency weighted IoU
        8. Kappa coefficient
        9. Matthews Correlation Coefficient
        # ----- for change detection task -----
        10. missed detection rate
        11. false alarm rate
        12. overall error
    """

    def __init__(self, num_class):

        self.num_class = num_class
        # confusion matrix: num_class*num_class
        """
        L\P      prediction
                 0      1
        label 0  TN    FP
              1  FN    TP
        """
        self.confusion_matrix = np.zeros((num_class,) * 2).astype('int64')
        self.label_range = range(self.num_class)

        self.eps = 1e-10

    def reset_confusion_matrix(self):
        # after computing the confusion matrix for each iter, then reset the confusion matrix
        self.confusion_matrix = np.zeros((self.num_class, )*2)

    def cal_confusion_matrix(self, input, label):
        """
        Args:
            input: model prediction (Type: numpy array)
            label: reference ground truth (Type: numpy array)
        Return:
            confusion matrix (Type: numpy array)

        """
        if label.max() != self.num_class - 1:
            raise ValueError('The value range in input should be consistent with the number of classes. Got {}'.format(label.max()))
        assert len(input.shape) == len(label.shape)
        input = input.astype('int')
        label = label.astype('int')

        mask = (label >= 0) & (label < self.num_class)
        label = self.num_class * label[mask] + input[mask]
        count = np.bincount(label, minlength=self.num_class ** 2)
        conf_mat = count.reshape(self.num_class, self.num_class)

        return conf_mat

    def cal_precision(self):
        # precision = TP / (TP + FP)
        precision = np.diag(self.confusion_matrix) / (np.sum(self.confusion_matrix, axis=0) + self.eps)
        return precision

    def cal_recall(self):
        # recall = TP / (TP + FN)
        recall = np.diag(self.confusion_matrix) / (np.sum(self.confusion_matrix, axis=1) + self.eps)
        return recall

    def cal_fbeta_score(self, beta=1):
        """
        Fbeta = (1+beta^2)*precision*recall/(beta^2*precision + recall)
        """
        pre = self.cal_precision()
        rec = self.cal_recall()
        f1 = 2 * beta * pre * rec / (beta * pre + rec + self.eps)
        return f1

    def cal_overall_accuracy(self):
        # acc = (TP + TN) / (TP + TN + FP + TN)
        acc = np.diag(self.confusion_matrix).sum() / self.confusion_matrix.sum()
        return acc

    def cal_IoU(self):
        # IoU = TP / (TP + FP + FN)
        intersection = np.diag(self.confusion_matrix)
        union = np.sum(self.confusion_matrix, axis=1) + np.sum(self.confusion_matrix, axis=0) - np.diag(self.confusion_matrix)
        IoU = intersection / (union + self.eps)
        return IoU

    def cal_mIoU(self):
        # Calculate mean IoU
        return np.nanmean(self.cal_IoU())

    def cal_fwIoU(self):
        # fwIoU = [(TP+FN)/(TP+FP+TN+FN)] *[TP / (TP + FP + FN)]
        freq = np.sum(self.confusion_matrix, axis=1) / np.sum(self.confusion_matrix)
        iu = np.diag(self.confusion_matrix) / (np.sum(self.confusion_matrix, axis=1) + np.sum(self.confusion_matrix, axis=0) -
                                               np.diag(self.confusion_matrix))
        fwIoU = (freq[freq > 0] * iu[freq > 0]).sum()
        return fwIoU

    def cal_Kappa(self):
        p0 = self.cal_overall_accuracy()
        cols = np.sum(self.confusion_matrix, axis=0)
        rows = np.sum(self.confusion_matrix, axis=1)
        sum_total = self.confusion_matrix.sum()
        Pe = np.dot(cols, rows) / float(sum_total ** 2)
        kappa = (p0 - Pe) / (1 - Pe + self.eps)

        return kappa

    def cal_MCC(self):
        tp = np.diag(self.confusion_matrix)

        # calculate tn
        tn = np.zeros(self.num_class)
        for i in range(self.num_class):
            temp = np.ma.masked_array(self.confusion_matrix, mask=False)
            temp.mask[i, :] = True
            temp.mask[:, i] = True
            tn[i] = temp.sum()

        fp = np.sum(self.confusion_matrix, axis=1) - tp
        fn = np.sum(self.confusion_matrix, axis=0) - tp

        # normalization
        n = self.confusion_matrix.sum()
        tn = tn / n
        fn = fn / n
        tp = tp / n
        fp = fp / n

        num = np.multiply(tp, tn) - np.multiply(fp, fn)
        den = np.multiply(tp + fp, tp + fn)
        den = np.multiply(den, tn + fp)
        den = np.multiply(den, tn + fn)
        den = np.sqrt(den)

        mcc = num / (den + self.eps)

        return mcc

    def cal_missed_detection_rate(self):
        """MD = Nm / Nc
           Nm = FN
        """
        fn = np.sum(self.confusion_matrix, axis=0) - np.diag(self.confusion_matrix)
        fn = fn.tolist()
        fn.reverse()
        fn = np.asarray(fn)
        p = np.sum(self.confusion_matrix, axis=1)
        md = fn / (p + self.eps)
        return md

    def cal_false_alarm_rate(self):
        """FA = FP / (TN + FP)"""
        fp = np.sum(self.confusion_matrix, axis=1) - np.diag(self.confusion_matrix)
        n = np.sum(self.confusion_matrix, axis=1)
        fa = fp / n
        fa = fa.tolist()
        fa.reverse()
        fa = np.asarray(fa)
        return fa

    def cal_overall_error(self):
        oa = self.cal_overall_accuracy()
        oe = 1 - oa
        return oe



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-31 15:55 
Written by Yuwei Jin (642281525@qq.com)
"""

import math
import torch
import torch.nn as nn
import torch.nn.functional as F

BatchNorm2d = nn.BatchNorm2d
bn_mom = 0.1


def conv3x3(in_planes, out_planes, stride=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, no_relu=False):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = BatchNorm2d(planes, momentum=bn_mom)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = BatchNorm2d(planes, momentum=bn_mom)
        self.downsample = downsample
        self.stride = stride
        self.no_relu = no_relu

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual

        if self.no_relu:
            return out
        else:
            return self.relu(out)


class Bottleneck(nn.Module):
    expansion = 2

    def __init__(self, inplanes, planes, stride=1, downsample=None, no_relu=False):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes, momentum=bn_mom)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes, momentum=bn_mom)
        self.conv3 = nn.Conv2d(planes, planes * self.expansion, kernel_size=1,
                               bias=False)
        self.bn3 = nn.BatchNorm2d(planes * self.expansion,
                                  momentum=bn_mom)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride
        self.no_relu = no_relu

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        if self.no_relu:
            return out
        else:
            return self.relu(out)


class DualResNet(nn.Module):

    def __init__(self, block, layers, num_classes=1000, planes=64, last_planes=2048):
        super(DualResNet, self).__init__()

        # self.inplanes = 64
        # fuse_planes = 128
        highres_planes = planes * 2
        self.last_planes = last_planes

        self.conv1 = nn.Sequential(
            nn.Conv2d(3, planes, kernel_size=3, stride=2, padding=1),
            BatchNorm2d(planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(planes, planes, kernel_size=3, stride=2, padding=1),
            BatchNorm2d(planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
        )

        self.relu = nn.ReLU(inplace=False)
        self.layer1 = self._make_layer(block, planes, planes, layers[0])
        self.layer2 = self._make_layer(block, planes, planes * 2, layers[1], stride=2)
        self.layer3 = self._make_layer(block, planes * 2, planes * 4, layers[2], stride=2)
        self.layer4 = self._make_layer(block, planes * 4, planes * 8, layers[3], stride=2)

        self.compression3 = nn.Sequential(
            nn.Conv2d(planes * 4, highres_planes, kernel_size=1, bias=False),
            BatchNorm2d(highres_planes, momentum=bn_mom),
        )

        self.compression4 = nn.Sequential(
            nn.Conv2d(planes * 8, highres_planes, kernel_size=1, bias=False),
            BatchNorm2d(highres_planes, momentum=bn_mom),
        )

        self.down3 = nn.Sequential(
            nn.Conv2d(highres_planes, planes * 4, kernel_size=3, stride=2, padding=1, bias=False),
            BatchNorm2d(planes * 4, momentum=bn_mom),
        )

        self.down4 = nn.Sequential(
            nn.Conv2d(highres_planes, planes * 4, kernel_size=3, stride=2, padding=1, bias=False),
            BatchNorm2d(planes * 4, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(planes * 4, planes * 8, kernel_size=3, stride=2, padding=1, bias=False),
            BatchNorm2d(planes * 8, momentum=bn_mom),
        )

        self.layer3_ = self._make_layer(block, planes * 2, highres_planes, 2)

        self.layer4_ = self._make_layer(block, highres_planes, highres_planes, 2)

        self.layer5_ = self._make_layer(Bottleneck, highres_planes, highres_planes, 1)

        self.down5 = nn.Sequential(
            nn.Conv2d(planes * 4, planes * 8, kernel_size=3, stride=2, padding=1, bias=False),
            BatchNorm2d(planes * 8, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.Conv2d(planes * 8, planes * 16, kernel_size=3, stride=2, padding=1, bias=False),
            BatchNorm2d(planes * 16, momentum=bn_mom),
        )

        self.layer5 = self._make_layer(Bottleneck, planes * 8, planes * 8, 1)

        self.last_layer = nn.Sequential(
            nn.Conv2d(planes * 16, last_planes, kernel_size=1, stride=1, padding=0, bias=False),
            BatchNorm2d(last_planes, momentum=bn_mom),
            nn.ReLU(inplace=True),
            nn.AdaptiveAvgPool2d((1, 1)),
        )

        self.linear = nn.Linear(last_planes, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion, momentum=bn_mom),
            )

        layers = []
        layers.append(block(inplanes, planes, stride, downsample))
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            if i == (blocks - 1):
                layers.append(block(inplanes, planes, stride=1, no_relu=True))
            else:
                layers.append(block(inplanes, planes, stride=1, no_relu=False))

        return nn.Sequential(*layers)

    def forward(self, x):

        width_output = x.shape[-1] // 8
        height_output = x.shape[-2] // 8
        layers = []

        x = self.conv1(x)

        x = self.layer1(x)
        layers.append(x)

        x = self.layer2(self.relu(x))
        layers.append(x)

        x = self.layer3(self.relu(x))
        layers.append(x)
        x_ = self.layer3_(self.relu(layers[1]))

        x = x + self.down3(self.relu(x_))
        x_ = x_ + F.interpolate(
            self.compression3(self.relu(layers[2])),
            size=[height_output, width_output],
            mode='bilinear')

        x = self.layer4(self.relu(x))
        layers.append(x)


        x_ = self.layer4_(self.relu(x_))

        x = x + self.down4(self.relu(x_))
        x_ = x_ + F.interpolate(
            self.compression4(self.relu(layers[3])),
            size=[height_output, width_output],
            mode='bilinear')

        x = self.layer5(self.relu(x)) + self.down5(self.relu(self.layer5_(self.relu(x_))))

        x = self.last_layer(self.relu(x))
        x = x.view(-1, self.last_planes)
        x = self.linear(x)
        return x


def get_model(pretrained=True, ckpt_src=None):
    model = DualResNet(block=BasicBlock, layers=[2, 2, 2, 2], planes=32, last_planes=1024)
    if pretrained:
        checkpoint = torch.load(ckpt_src, map_location='cpu')
        model.load_state_dict(checkpoint)
    return model


if __name__ == '__main__':
    model = get_model()
    model(torch.randn(1, 3, 224, 224))


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-05-04 16:47
Written by Yuwei Jin (642281525@qq.com)
"""
import os

"""
resnet model, borrowed from EMANet. Thanks the authors for providing their codes.
"""

import math
import torch
import torch.nn as nn

from sv.libs import init_weights

norm_layer = nn.BatchNorm2d


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, dilation=1, downsample=None):
        super().__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, 1, bias=False)
        self.bn1 = norm_layer(planes)
        self.conv2 = nn.Conv2d(planes, planes, 3, stride, dilation, dilation, bias=False)
        self.bn2 = norm_layer(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, 1, bias=False)
        self.bn3 = norm_layer(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.dilation = dilation
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class ResNet(nn.Module):
    def __init__(self, block, layers, num_classes=1000, stride=8):
        self.inplanes = 128
        super().__init__()

        self.conv1 = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=3, stride=2, padding=1, bias=False),
            norm_layer(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=1, bias=False),
            norm_layer(64),
            nn.ReLU(inplace=True),
            nn.Conv2d(64, 128, kernel_size=3, stride=1, padding=1, bias=False)
        )
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)

        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)

        self.layer1 = self._make_layer(block, 64, layers[0])
        self.layer2 = self._make_layer(block, 128, layers[1], stride=2)

        if stride == 16:
            self.layer3 = self._make_layer(block, 256, layers[2], stride=2)
            self.layer4 = self._make_layer(block, 512, layers[3], stride=1, dilation=2, grids=[1, 2, 4])
        elif stride == 8:
            self.layer3 = self._make_layer(block, 256, layers[2], stride=1, dilation=2)
            self.layer4 = self._make_layer(block, 512, layers[3], stride=1, dilation=4, grids=[1, 2, 4])
        else:
            self.layer3 = self._make_layer(block, 256, layers[2], stride=2, dilation=1)
            self.layer4 = self._make_layer(block, 512, layers[3], stride=2, dilation=1)

        self.avgpool = nn.AvgPool2d(7, stride=1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                if m.bias is not None:
                    m.bias.data.zero_()

    def _make_layer(self, block, planes, blocks, stride=1, dilation=1, grids=None):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion, kernel_size=1, stride=stride, bias=False),
                norm_layer(planes * block.expansion)
            )

        layers = []
        if grids is None:
            grids = [1] * blocks

        if dilation == 1 or dilation == 2:
            layers.append(block(self.inplanes, planes, stride, dilation=1, downsample=downsample))
        elif dilation == 4:
            layers.append(block(self.inplanes, planes, stride, dilation=2, downsample=downsample))
        else:
            raise RuntimeError('=> unknown dilation size: {}'.format(dilation))

        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, dilation=dilation * grids[i]))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x0 = self.relu(x)
        x = self.maxpool(x0)

        x1 = self.layer1(x)
        x2 = self.layer2(x1)
        x3 = self.layer3(x2)
        x4 = self.layer4(x3)

        return x1, x2, x3, x4


def resnet(n_layers, stride, ckpt_folder, multi_grid=False):
    supported_model = [50, 101]
    if n_layers in supported_model:
        layers = {
            50: [3, 4, 6, 3],
            101: [3, 4, 23, 3]
        }[n_layers]
        pretrained_path = {
            50: os.path.join(ckpt_folder, 'resnet50-ebb6acbb.pth'),
            101: os.path.join(ckpt_folder, 'resnet101-2a57e44d.pth')
        }[n_layers]

        net = ResNet(Bottleneck, layers=layers, stride=stride)
        print("loading resnet pretrained weight....")
        state_dict = torch.load(pretrained_path)
        net.load_state_dict(state_dict, strict=False)
        print("loaded pretrained imagenet weight successfully.")

        if multi_grid:
            multi_grid_dilations = [1, 2, 4]
            net.layer4.add_module('layer5', nn.Sequential(
                Bottleneck(inplanes=2048, planes=512, dilation=2 * multi_grid_dilations[0]),
                Bottleneck(inplanes=2048, planes=512, dilation=2 * multi_grid_dilations[1]),
                Bottleneck(inplanes=2048, planes=512, dilation=2 * multi_grid_dilations[2])
            ))
            net.layer4.add_module('layer6', nn.Sequential(
                Bottleneck(inplanes=2048, planes=512, dilation=4 * multi_grid_dilations[0]),
                Bottleneck(inplanes=2048, planes=512, dilation=4 * multi_grid_dilations[1]),
                Bottleneck(inplanes=2048, planes=512, dilation=4 * multi_grid_dilations[2])
            ))
            net.layer4.add_module('layer7', nn.Sequential(
                Bottleneck(inplanes=2048, planes=512, dilation=8 * multi_grid_dilations[0]),
                Bottleneck(inplanes=2048, planes=512, dilation=8 * multi_grid_dilations[1]),
                Bottleneck(inplanes=2048, planes=512, dilation=8 * multi_grid_dilations[2])
            ))

            init_weights(net.layer4.layer5, net.layer4.layer6, net.layer4.layer7)

        return net
    else:
        raise NotImplementedError


if __name__ == '__main__':
    model = resnet(50, 32, multi_grid=False)

    from torchstat import stat
    stat(model, (3, 512, 512))
    # x = torch.randn(3, 3, 256, 256)
    # y = model(x)
    # z = 0

    print()



# -*- coding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2021-03-10 20:01
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
import torch.nn as nn

from sv.libs.tools.init_weight import init_weights


class CAM(nn.Module):
    def __init__(self, in_chs, ratio=16):
        super(CAM, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.max_pool = nn.AdaptiveMaxPool2d(1)

        self.fc1 = nn.Conv2d(in_chs, in_chs // 16, 1, bias=False)
        self.relu1 = nn.ReLU()
        self.fc2 = nn.Conv2d(in_chs // 16, in_chs, 1, bias=False)

        self.sigmoid = nn.Sigmoid()

        init_weights(self.fc1, self.fc2)

    def forward(self, x):
        avg_out = self.fc2(self.relu1(self.fc1(self.avg_pool(x))))
        max_out = self.fc2(self.relu1(self.fc1(self.max_pool(x))))
        y = avg_out + max_out
        out = self.sigmoid(y) * x

        return out


class SAM(nn.Module):
    def __init__(self):
        super(SAM, self).__init__()
        self.conv1 = nn.Conv2d(2, 1, kernel_size=7, padding=3, bias=False)
        self.sigmoid = nn.Sigmoid()

        init_weights(self.conv1)

    def forward(self, x):
        avg_out = torch.mean(x, dim=1, keepdim=True)
        max_out, _ = torch.max(x, dim=1, keepdim=True)
        y = torch.cat([avg_out, max_out], dim=1)
        y = self.conv1(y)
        y = self.sigmoid(y)
        out = y * x

        return out


class CBAM(nn.Module):
    def __init__(self, in_chs):
        super(CBAM, self).__init__()

        self.cam = CAM(in_chs)
        self.sam = SAM()

    def forward(self, x):
        x = self.cam(x)
        y = self.sam(x)

        return y



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-30 18:56 
Written by Yuwei Jin (642281525@qq.com)
"""

import torch
import torch.nn as nn

from torchvision.models.mobilenetv3 import mobilenet_v3_small


class MobileNetv3(nn.Module):
    def __init__(self):
        super(MobileNetv3, self).__init__()

        net = mobilenet_v3_small(pretrained=True)

        self.layer1 = net.features[:2]  # 4x
        self.layer2 = net.features[2:4]  # 8x
        self.layer3 = net.features[4:9]  # 16x
        self.layer4 = net.features[9:11]  # 32x

    def forward(self, x):
        y1 = self.layer1(x)
        y2 = self.layer2(y1)
        y3 = self.layer3(y2)
        y4 = self.layer4(y3)
        return y1, y2, y3, y4


if __name__ == '__main__':
    model = MobileNetv3()

    from torchstat import stat

    stat(model, (3, 512, 512))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-05-19 14:21 
Written by Yuwei Jin (642281525@qq.com)
"""
import logging
import os
from collections import OrderedDict

import torch.nn as nn


class BN_AC_Conv2d(nn.Module):

    def __init__(self, num_in, num_filter, kernel=1, pad=0, stride=1, g=1, bias=False):
        super().__init__()
        self.bn = nn.BatchNorm2d(num_in)
        self.relu = nn.ReLU(inplace=True)
        self.conv = nn.Conv2d(num_in, num_filter, kernel_size=kernel, padding=pad, stride=stride, groups=g, bias=bias)

    def forward(self, x):
        h = self.relu(self.bn(x))
        h = self.conv(h)
        return h


class MF_UNIT(nn.Module):

    def __init__(self, num_in, num_mid, num_out, g=1, stride=1, first_block=False):
        super(MF_UNIT, self).__init__()
        num_ix = int(num_mid / 4)

        # prepare input
        self.conv_i1 = BN_AC_Conv2d(num_in=num_in, num_filter=num_ix, kernel=1, pad=0)
        self.conv_i2 = BN_AC_Conv2d(num_in=num_ix, num_filter=num_in, kernel=1, pad=0)

        # main part
        self.conv_m1 = BN_AC_Conv2d(num_in=num_in, num_filter=num_mid, kernel=3, pad=1, stride=stride, g=g)

        if first_block:
            self.conv_m2 = BN_AC_Conv2d(num_in=num_mid, num_filter=num_out, kernel=1, pad=0)
        else:
            self.conv_m2 = BN_AC_Conv2d(num_in=num_mid, num_filter=num_out, kernel=3, pad=1, g=g)

        # adapter
        if first_block:
            self.conv_w1 = BN_AC_Conv2d(num_in=num_in, num_filter=num_out, kernel=1, pad=0, stride=stride)

    def forward(self, x):

        h = self.conv_i1(x)
        x_in = x + self.conv_i2(h)

        h = self.conv_m1(x_in)
        h = self.conv_m2(h)

        if hasattr(self, 'conv_w1'):
            x = self.conv_w1(x)

        return h + x


class MFNET_2D(nn.Module):

    def __init__(self, num_classes=1000, pretrained=False, groups=16, ckpt_src=None):
        super(MFNET_2D, self).__init__()

        k_sec = {2: 3,
                 3: 4,
                 4: 6,
                 5: 3}

        # conv1 - x224 (x16)
        conv1_num_out = 16
        self.conv1 = nn.Sequential(
            OrderedDict([
                ('conv', nn.Conv2d(3, conv1_num_out, kernel_size=5, padding=2, stride=2, bias=False)),
                ('bn', nn.BatchNorm2d(conv1_num_out)),
                ('relu', nn.ReLU(inplace=True))
            ])
        )
        # self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.maxpool = nn.Sequential(
            nn.Conv2d(16, 32, 3, padding=1, stride=2),
            nn.BatchNorm2d(32),
            nn.ReLU(True)
        )

        # conv2 - x56 (x8)
        num_mid = 96
        conv2_num_out = 96
        self.conv2 = nn.Sequential(
            OrderedDict([
                ("B%02d" % i,
                 MF_UNIT(num_in=32 if i == 1 else conv2_num_out,
                         num_mid=num_mid,
                         num_out=conv2_num_out,
                         stride=1,
                         g=groups,
                         first_block=(i==1))
                 ) for i in range(1, k_sec[2] + 1)
            ])
        )

        # conv3 - x28 (x8)
        num_mid *= 2
        conv3_num_out = 2 * conv2_num_out
        self.conv3 = nn.Sequential(OrderedDict([
            ("B%02d" % i, MF_UNIT(num_in=conv2_num_out if i == 1 else conv3_num_out,
                                  num_mid=num_mid,
                                  num_out=conv3_num_out,
                                  stride=2 if i == 1 else 1,
                                  g=groups,
                                  first_block=(i==1))
             ) for i in range(1, k_sec[3] + 1)
        ]))


        # conv4 - x14 (x8)
        num_mid *= 2
        conv4_num_out = 2 * conv3_num_out
        self.conv4 = nn.Sequential(OrderedDict([
            ("B%02d" % i, MF_UNIT(num_in=conv3_num_out if i == 1 else conv4_num_out,
                                  num_mid=num_mid,
                                  num_out=conv4_num_out,
                                  stride=2 if i == 1 else 1,
                                  g=groups,
                                  first_block=(i == 1))
             ) for i in range(1, k_sec[4] + 1)
        ]))

        # conv5 - x7 (x8)
        num_mid *= 2
        conv5_num_out = 2 * conv4_num_out
        self.conv5 = nn.Sequential(OrderedDict([
            ("B%02d" % i, MF_UNIT(num_in=conv4_num_out if i == 1 else conv5_num_out,
                                  num_mid=num_mid,
                                  num_out=conv5_num_out,
                                  stride=2 if i == 1 else 1,
                                  g=groups,
                                  first_block=(i == 1))
             ) for i in range(1, k_sec[5] + 1)
        ]))

        # final
        self.tail = nn.Sequential(OrderedDict([
            ('bn', nn.BatchNorm2d(conv5_num_out)),
            ('relu', nn.ReLU(inplace=True))
        ]))

        self.globalpool = nn.Sequential(OrderedDict([
            ('avg', nn.AvgPool3d(kernel_size=7, stride=1)),
            # ('dropout', nn.Dropout(p=0.5)), only for fine-tuning
        ]))
        self.classifier = nn.Linear(conv5_num_out, num_classes)

        #############
        # Initialization

        if pretrained:
            import torch
            load_method = 'inflation'  # 'random', 'inflation'
            pretrained_model = ckpt_src
            logging.info("Network:: graph initialized, loading pretrained model: `{}'".format(pretrained_model))
            assert os.path.exists(pretrained_model), "cannot locate: `{}'".format(pretrained_model)
            state_dict_2d = torch.load(pretrained_model)
            self.load_state_dict(state_dict_2d, strict=True)
            logging.info("loading pretrained model: `{}' successfully".format(pretrained_model))
        else:
            logging.info("Network:: graph initialized, use random inilization!")

    def forward(self, x):
        # assert x.shape[2] == 16

        h = self.conv1(x)  # x224 -> x112
        h = self.maxpool(h)  # x112 ->  x56

        h1 = self.conv2(h)  # x56 ->  x56
        h2 = self.conv3(h1)  # x56 ->  x28
        h3 = self.conv4(h2)  # x28 ->  x14
        h4 = self.conv5(h3)  # x14 ->   x7

        # h = self.tail(h)
        # h = self.globalpool(h)
        #
        # h = h.view(h.shape[0], -1)
        # h = self.classifier(h)

        return h1, h2, h3, h4


if __name__ == "__main__":
    import torch

    logging.getLogger().setLevel(logging.DEBUG)
    # ---------
    net = MFNET_2D(num_classes=1000, pretrained=True)
    data = torch.autograd.Variable(torch.randn(1, 3, 224, 224))
    output = net(data)
    torch.save({'state_dict': net.state_dict()}, './tmp.pth')

    del net.classifier, net.tail, net.globalpool

    from torchstat import stat
    stat(net, (3, 512, 512))



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-06-13 20:14
Written by Yuwei Jin (642281525@qq.com)
"""

import json
import cv2
import os
import numpy as np
from time import time
from multiprocessing import Pool

# from configs import cfg
# from tools.metrics import BETestMetrics
# from tools import utils
from sv.libs.tools import utils


def _get_big_im_names(cfg):
    im_names = []
    for f in os.listdir(cfg.DATASET.TEST_WHOLE_LABEL):
        if f.endswith(cfg.DATASET.IMG_EXTENSION):
            temp, _ = f.split('.')
            im_names.append(temp)
    return im_names


def _load_crop_info(file, cfg):
    src = os.path.join(cfg.DATASET.TEST_SET, 'crop_info', file + '_crop_info' + '.json')
    with open(src) as json_file:
        info = json.load(json_file)

    rows = info['rows']
    cols = info['cols']
    patch_size = info['patch_size']
    m = info['m']
    n = info['n']

    if 'overlap' in info.keys():
        overlap = info['overlap']
        return rows, cols, patch_size, overlap, m, n
    else:
        central_win_size = info['central_win_size']
        return rows, cols, patch_size, central_win_size, m, n


def _read_prob_img(img_name: str, index: int, cfg):
    # name format: model name + '_' + image name + '_prob.png'
    temp = cfg.MODEL.NAME + '_' + img_name + '-' + str(index) + '_prob' + '.png'

    src = os.path.join(cfg.TEST.PROB_SAVE_DIR, temp)
    prob = cv2.imread(src, cv2.IMREAD_GRAYSCALE) / 255  # value should be [0,1]

    return prob


class Stitch:
    def __init__(self, metrics, cfg):
        self.evaluator = metrics()
        utils.mk_dirs_r(cfg.TEST.STITCH_SAVE_DIR)
        self.cfg = cfg

    def _save_res(self, im_name, whole_prob, gt):
        self.evaluator.cal_batch_metric(im_name, whole_prob, gt)
        utils.save_prob_map(im_name, whole_prob, self.cfg, self.cfg.TEST.STITCH_SAVE_DIR)
        utils.save_alpha_map(im_name, whole_prob, gt, self.cfg, self.cfg.TEST.STITCH_SAVE_DIR)

    def _sub_func_dilated(self, im_name):
        d = self.cfg.DATASET
        print("stitching: {}{}".format(im_name, d.IMG_EXTENSION))

        rows, cols, patch_size, win, m, n = _load_crop_info(im_name, self.cfg)
        pad = int((patch_size - win) // 2)

        whole_prob = np.zeros((rows, cols))

        k = 0
        for i in range(m):
            x1 = i * win
            x2 = x1 + win
            if i == m - 1:  # 超出图像范围
                x1, x2 = (m - 1) * win, rows

            for j in range(n):
                y1 = j * win
                y2 = y1 + win
                if j == n - 1:  # 超出图像范围
                    y1, y2 = (n - 1) * win, cols

                k += 1

                patch = _read_prob_img(im_name, index=k)

                a1, a2 = pad, pad + win
                b1, b2 = pad, pad + win
                if i == m - 1:
                    a1, a2 = pad, pad + rows - (m - 1) * win
                if j == n - 1:
                    b1, b2 = pad, pad + cols - (n - 1) * win

                # weight[x1:x2, y1:y2] = weight[x1:x2, y1:y2] + 1
                whole_prob[x1:x2, y1:y2] = patch[a1:a2, b1:b2]

        gt_src = os.path.join(d.TEST_WHOLE_LABEL, im_name + d.IMG_EXTENSION)
        gt = cv2.imread(gt_src, cv2.IMREAD_GRAYSCALE) / 255

        self.evaluator.cal_batch_metric(im_name, whole_prob, gt)

        if self.cfg.TEST.STITCH_SAVE_RES:
            self._save_res(whole_prob=whole_prob, im_name=im_name, gt=gt)

    def _sub_func_overlap(self, im_name):
        d = self.cfg.DATASET
        print("stitching: {}{}".format(im_name, d.IMG_EXTENSION))

        rows, cols, patch_size, overlap, m, n = _load_crop_info(im_name)
        weight = np.zeros((rows, cols))
        whole_prob = np.zeros((rows, cols))

        k = 0
        for i in range(m):
            x1 = i * (patch_size - overlap)
            x2 = x1 + patch_size
            if x2 >= rows:  # 超出图像范围
                x1, x2 = rows - patch_size, rows
            for j in range(n):
                y1 = j * (patch_size - overlap)
                y2 = y1 + patch_size
                if y2 >= cols:  # 超出图像范围
                    y1, y2 = cols - patch_size, cols

                k += 1

                patch = _read_prob_img(img_name=im_name, index=k)

                weight[x1:x2, y1:y2] = weight[x1:x2, y1:y2] + 1
                whole_prob[x1:x2, y1:y2] = whole_prob[x1:x2, y1:y2] + patch

        whole_prob = whole_prob / weight

        gt_src = os.path.join(d.TEST_WHOLE_LABEL, im_name + d.IMG_EXTENSION)
        gt = cv2.imread(gt_src, cv2.IMREAD_GRAYSCALE) / 255

        self.evaluator.cal_batch_metric(im_name, whole_prob, gt)

        if self.cfg.TEST.STITCH_SAVE_RES:
            self._save_res(whole_prob=whole_prob, im_name=im_name, gt=gt)

    def sub_func(self, func):
        """func: specific stitch function"""
        tic = time()
        files = _get_big_im_names()
        # if len(files) >= 8:  # applying multi processing
        #     pool = Pool(processes=8)
        #     pool.map(func, files)
        #     pool.close()
        #     pool.join()
        # else:
        for f in files:
            func(f)
        toc = time()
        print("Stitch processing elapsed time: {:.2f}".format(toc - tic) + "s")
        self.evaluator.average()

    def run(self):
        if self.cfg.TEST.USE_DILATED_PREDICT:
            self.sub_func(self._sub_func_dilated)
        else:
            self.sub_func(self._sub_func_overlap)



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-05-13 14:10
Written by Yuwei Jin (642281525@qq.com)
"""

import torch
import warnings
from tqdm import tqdm

from sv.libs.core.infer import Infer
from sv.libs.data.dataset.dataset import load_test_data
from sv.libs.tools import utils

warnings.filterwarnings("ignore")

# load well-trained model weights from disk.


class Test(Infer):
    def __init__(self, model, metrics, cfg):
        super(Test, self).__init__(model, cfg)

        self.evaluator = metrics

        self.test_data = load_test_data(cfg)
        self.cfg = cfg

    def test(self):
        with torch.no_grad():
            for i, (im_name, im, gt) in enumerate(tqdm(self.test_data, ncols=80, desc='inferring')):
                prob = self.forward(im)

                im_name, _ = im_name[-1].split('.')
                if isinstance(prob, (list, tuple)):
                    prob = prob[0]

                self.evaluator.cal_batch_metric(im_name, prob, gt)

                if self.cfg.TEST.SAVE_RES:
                    utils.save_prob_map(im_name, prob, self.cfg)
                    if not self.cfg.TEST.STITCH_RES:
                        utils.save_alpha_map(im_name, prob, gt, self.cfg)

        self.evaluator.average()


# def test():
#     # if cfg.TEST.SAVE_RES:
#     #     utils.mk_dirs_r(cfg.TEST.PROB_SAVE_DIR)
#     #
#     # Test()()
#     if cfg.TEST.STITCH_RES:
#         from tools.stitch_probs import Stitch
#         Stitch()()

#
# if __name__ == '__main__':
#     test()



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-24 16:45 
Written by Yuwei Jin (642281525@qq.com)
"""

from setuptools import find_packages, setup

setup(
    name="sv_seg",
    version="1.1",
    description="A simplified version library for deep learning in semantic segmentation",
    author="jyw",
    license='MIT',
    repository='https://upload.pypi.org/legacy/',
    packages=find_packages(),
    classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'Topic :: System :: Logging',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
        ],
    # py_modules=['config_samples', 'libs', 'modules']
)



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 17:20 
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
import torch.nn as nn

from sv.libs.tools.init_weight import init_weights


# ------ 3x3 Conv -----
class Conv3x3(nn.Module):
    def __init__(self, in_chs, out_chs, stride=1, dilation=1, groups=1, bias=True):
        super(Conv3x3, self).__init__()
        self.conv = nn.Conv2d(in_chs, out_chs, kernel_size=3, padding=dilation, stride=stride, dilation=dilation, groups=groups, bias=bias)
        init_weights(self.conv)

    def forward(self, x):
        out = self.conv(x)
        return out


# ------ 3x3 Conv + BN + ReLU ------
class Conv3x3_BN_ReLU(nn.Module):
    def __init__(self, in_chs, out_chs, stride=1, padding=1, dilation=1, groups=1, bn_momentum=0.1):
        super(Conv3x3_BN_ReLU, self).__init__()

        if dilation > 1:
            padding = dilation

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=3, stride=stride, padding=padding, dilation=dilation,
                      groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.ReLU(inplace=True)
        )

        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


class Conv3x3_BN_PReLU(nn.Module):
    def __init__(self, in_chs, out_chs, stride=1, padding=1, dilation=1, groups=1, bn_momentum=0.1):
        super().__init__()

        if dilation > 1:
            padding = dilation

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=3, stride=stride, padding=padding, dilation=dilation,
                      groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.PReLU()
        )

        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)

class Conv3x3_BN_LeakyReLU(nn.Module):
    def __init__(self, in_chs, out_chs, stride=1, padding=1, dilation=1, groups=1, bn_momentum=0.1, negative_slope=0.1):
        super().__init__()

        if dilation > 1:
            padding = dilation

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=3, stride=stride, padding=padding, dilation=dilation,
                      groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True)
        )

        init_weights(self.conv, nonlinearity='leakyrelu')

    def forward(self, x):
        return self.conv(x)


class BN_ReLU_Conv3x3(nn.Module):
    """ pre-activate """

    def __init__(self, in_chs, out_chs, stride=1, padding=1, dilation=1, groups=1, momentum=0.1):
        super(BN_ReLU_Conv3x3, self).__init__()

        self.conv = nn.Sequential(
            nn.BatchNorm2d(in_chs, momentum=momentum),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_chs, out_chs, kernel_size=3, stride=stride, padding=padding, dilation=dilation, groups=groups, bias=False)
        )

        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


# ----- Conv 1x1 -----
class Conv1x1(nn.Module):
    def __init__(self, in_chs, out_chs, stride=1, groups=1, bias=True):
        super(Conv1x1, self).__init__()
        self.conv = nn.Conv2d(in_chs, out_chs, kernel_size=1, stride=stride, groups=groups, bias=bias)

        init_weights(self.conv)

    def forward(self, x):
        out = self.conv(x)
        return out


class Conv1x1_BN_ReLU(nn.Module):
    def __init__(self, in_chs, out_chs, groups=1, bn_momentum=0.1):
        super(Conv1x1_BN_ReLU, self).__init__()

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=1, groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.ReLU(inplace=True)
        )
        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


class Conv1x1_BN_PReLU(nn.Module):
    def __init__(self, in_chs, out_chs, groups=1, bn_momentum=0.1):
        super().__init__()

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=1, groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.PReLU()
        )
        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


class Conv1x1_BN_LeakyReLU(nn.Module):
    def __init__(self, in_chs, out_chs, groups=1, bn_momentum=0.1):
        super().__init__()

        self.conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=1, groups=groups, bias=False),
            nn.BatchNorm2d(out_chs, momentum=bn_momentum),
            nn.PReLU()
        )
        init_weights(self.conv, nonlinearity='leakyrelu')

    def forward(self, x):
        return self.conv(x)


class BN_ReLU_Conv1x1(nn.Module):
    """ pre-activate """

    def __init__(self, in_chs, out_chs, stride=1, dilation=1, groups=1, momentum=0.1):
        super(BN_ReLU_Conv1x1, self).__init__()

        self.conv = nn.Sequential(
            nn.BatchNorm2d(in_chs, momentum=momentum),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_chs, out_chs, kernel_size=1, stride=stride, dilation=dilation, groups=groups, bias=False)
        )

        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


class ConvBNReLU(nn.Module):
    def __init__(self, in_chan, out_chan, ks=3, stride=1, padding=1, *args, **kwargs):
        super(ConvBNReLU, self).__init__()
        self.conv = nn.Conv2d(in_chan, out_chan, kernel_size=ks, stride=stride, padding=padding, bias=False)
        # self.bn = BatchNorm2d(out_chan)
        self.bn = nn.BatchNorm2d(out_chan)
        self.relu = nn.ReLU()

        init_weights(self.conv, self.bn)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x



class DWConv(nn.Module):
    """Depthwise-separable convolution"""

    def __init__(self, in_chs, out_chs, dilation=1, relu=True):
        super(DWConv, self).__init__()

        self.depth_conv = nn.Sequential(
            nn.Conv2d(in_chs, in_chs, kernel_size=3, groups=in_chs, padding=dilation, dilation=dilation, bias=False),
            nn.BatchNorm2d(in_chs)
        )

        self.point_conv = nn.Sequential(
            nn.Conv2d(in_chs, out_chs, kernel_size=1, bias=False),
            nn.BatchNorm2d(out_chs)
        )

        self.relu = relu

        init_weights(self.point_conv, self.depth_conv)

    def forward(self, x):
        out = self.depth_conv(x)
        out = self.point_conv(out)

        if self.relu:
            out = torch.relu(out)

        return out


# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-15 10:49
Written by Yuwei Jin (642281525@qq.com)
"""

import torch

from torch.nn import Module, Conv2d, Softmax, Parameter

from sv.libs.tools.init_weight import init_weights


class PAM_Module(Module):
    """ Position attention module"""

    # Ref from SAGAN
    def __init__(self, in_dim):
        super(PAM_Module, self).__init__()
        self.chanel_in = in_dim

        self.query_conv = Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.key_conv = Conv2d(in_channels=in_dim, out_channels=in_dim // 8, kernel_size=1)
        self.value_conv = Conv2d(in_channels=in_dim, out_channels=in_dim, kernel_size=1)
        self.gamma = Parameter(torch.zeros(1))

        self.softmax = Softmax(dim=-1)

        init_weights(self.query_conv, self.key_conv, self.value_conv)

    def forward(self, x):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            returns :
                out : attention value + input feature
                attention: B X (HxW) X (HxW)
        """
        m_batchsize, C, height, width = x.size()
        proj_query = self.query_conv(x).view(m_batchsize, -1, width * height).permute(0, 2, 1)
        proj_key = self.key_conv(x).view(m_batchsize, -1, width * height)
        energy = torch.bmm(proj_query, proj_key)
        attention = self.softmax(energy)
        proj_value = self.value_conv(x).view(m_batchsize, -1, width * height)

        out = torch.bmm(proj_value, attention.permute(0, 2, 1))
        out = out.view(m_batchsize, C, height, width)

        out = self.gamma * out + x
        return out


class CAM_Module(Module):
    """ Channel attention module"""

    def __init__(self, in_dim):
        super(CAM_Module, self).__init__()
        self.chanel_in = in_dim

        self.gamma = Parameter(torch.zeros(1))
        self.softmax = Softmax(dim=-1)

    def forward(self, x):
        """
            inputs :
                x : input feature maps( B X C X H X W)
            returns :
                out : attention value + input feature
                attention: B X C X C
        """
        m_batchsize, C, height, width = x.size()
        proj_query = x.view(m_batchsize, C, -1)
        proj_key = x.view(m_batchsize, C, -1).permute(0, 2, 1)
        energy = torch.bmm(proj_query, proj_key)
        energy_new = torch.max(energy, -1, keepdim=True)[0].expand_as(energy) - energy
        attention = self.softmax(energy_new)
        proj_value = x.view(m_batchsize, C, -1)

        out = torch.bmm(attention, proj_value)
        out = out.view(m_batchsize, C, height, width)

        out = self.gamma * out + x
        return out


class DAM(Module):
    def __init__(self, in_chs):
        super(DAM, self).__init__()
        self.pam = PAM_Module(in_dim=in_chs)
        self.cam = CAM_Module(in_dim=in_chs)

    def forward(self, x):
        x1 = self.pam(x)
        x2 = self.cam(x)

        return x1 + x2



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-07-15 09:09 
Written by Yuwei Jin (642281525@qq.com)
"""
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.autograd import Variable


# from libs.functional import edge_extraction


class BinaryDiceLoss(nn.Module):
    """Dice loss of binary class
        Args:
            smooth: A float number to smooth loss, and avoid NaN error, default: 1
            p: Denominator value: \sum{x^p} + \sum{y^p}, default: 2
            predict: A tensor of shape [N, *]
            target: A tensor of shape same with predict
            reduction: Reduction method to apply, return mean over batch if 'mean',
                return sum if 'sum', return a tensor of shape [N,] if 'none'
        Returns:
            Loss tensor according to arg reduction
        Raise:
            Exception if unexpected reduction
        """

    def __init__(self, smooth=1., p=2, reduction='mean'):
        super(BinaryDiceLoss, self).__init__()
        self.smooth = smooth
        self.p = p
        self.reduction = reduction

    def forward(self, input, target):
        assert input.shape[0] == target.shape[0], "predict & target batch size don't match"

        input = torch.sigmoid(input)

        n = input.size()[0]

        prob = input.view(n, -1)
        mask = target.view(n, -1)
        intersection = (prob * mask).sum(1)
        loss = 1 - ((2. * intersection + self.smooth) / (prob.sum(1) + mask.sum(1) + self.smooth))

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        elif self.reduction == 'none':
            return loss
        else:
            raise Exception('Unexpected reduction {}'.format(self.reduction))


class BCEDiceLoss(nn.Module):
    def __init__(self, dice_weight=1.0):
        super(BCEDiceLoss, self).__init__()
        self.bce = nn.BCEWithLogitsLoss()
        self.dice = BinaryDiceLoss()
        self.dice_weight = dice_weight

    def forward(self, x, y) -> torch.Tensor:
        """
        :param x: model predictions
        :param y: label
        :return: loss
        """
        t = torch.sigmoid(x)

        loss = self.bce(x, y) + self.dice_weight * self.dice(t, y)

        return loss


class OHEM_CELoss(nn.Module):
    """OHEM Cross entropy loss"""

    def __init__(self, ignore_index: int = -1, thresh: float = 0.7, min_kept: int = 100000):
        super(OHEM_CELoss, self).__init__()

        self.ignore_index = ignore_index
        self.thresh = thresh
        self.min_kept = min_kept

    def forward(self, y_pred, y_true):
        # y_pred: [N, C, H, W]
        # y_true: [N, H, W]
        # seg_weight: [N, H, W]
        y_true = y_true.unsqueeze(1)
        with torch.no_grad():
            assert y_pred.shape[2:] == y_true.shape[2:]
            assert y_true.shape[1] == 1
            seg_label = y_true.squeeze(1).long()
            batch_kept = self.min_kept * seg_label.size(0)
            valid_mask = seg_label != self.ignore_index
            seg_weight = y_pred.new_zeros(size=seg_label.size())
            valid_seg_weight = seg_weight[valid_mask]

            seg_prob = F.softmax(y_pred, dim=1)

            tmp_seg_label = seg_label.clone().unsqueeze(1)
            tmp_seg_label[tmp_seg_label == self.ignore_index] = 0
            seg_prob = seg_prob.gather(1, tmp_seg_label).squeeze(1)
            sort_prob, sort_indices = seg_prob[valid_mask].sort()

            if sort_prob.numel() > 0:
                min_threshold = sort_prob[min(batch_kept,
                                              sort_prob.numel() - 1)]
            else:
                min_threshold = 0.0
            threshold = max(min_threshold, self.thresh)
            valid_seg_weight[seg_prob[valid_mask] < threshold] = 1.

            seg_weight[valid_mask] = valid_seg_weight

            losses = F.cross_entropy(y_pred, y_true.squeeze(1), ignore_index=self.ignore_index, reduction='none')
            losses = losses * seg_weight

            return losses.sum() / seg_weight.sum()


def _masked_ignore(y_pred: torch.Tensor, y_true: torch.Tensor, ignore_index: int):
    # usually used for BCE-like loss
    y_pred = y_pred.reshape((-1,))
    y_true = y_true.reshape((-1,))
    valid = y_true != ignore_index
    y_true = y_true.masked_select(valid).float()
    y_pred = y_pred.masked_select(valid).float()
    return y_pred, y_true


class LabelSmoothingCELoss(nn.Module):
    def __init__(self, eps: float = 0.1, reduction: str = 'mean', ignore_index: int = -1):
        super(LabelSmoothingCELoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.ignore_index = ignore_index

    def forward(self, output: torch.Tensor, target: torch.Tensor):
        c = output.size()[0]
        log_preds = F.log_softmax(output, dim=1)

        loss = -log_preds.sum(dim=1)

        loss, _ = _masked_ignore(loss, target, self.ignore_index)

        if self.reduction == 'sum':
            loss = loss.sum()
        elif self.reduction == 'mean':
            loss = loss.mean()

        nll_loss = F.nll_loss(log_preds, target, reduction=self.reduction, ignore_index=self.ignore_index)
        loss = loss * self.eps / c + (1 - self.eps) * nll_loss

        return loss


def label_smoothing_binary_cross_entropy(output: torch.Tensor, target: torch.Tensor, eps: float = 0.1,
                                         reduction: str = 'mean', ignore_index: int = 255):
    output, target = _masked_ignore(output, target, ignore_index)
    target = torch.where(target == 0, target + eps, target - eps)
    return F.binary_cross_entropy_with_logits(output, target, reduction=reduction)


def binary_cross_entropy_with_logits(output: torch.Tensor, target: torch.Tensor, reduction: str = 'mean',
                                     ignore_index: int = 255):
    output, target = _masked_ignore(output, target, ignore_index)
    return F.binary_cross_entropy_with_logits(output, target, reduction=reduction)


@torch.jit.script
def online_hard_example_mining(losses: torch.Tensor, keep_ratio: float):
    assert 0 < keep_ratio < 1, 'The value of keep_ratio must be from 0 to 1.'
    # 1. keep num
    num_inst = losses.numel()
    num_hns = int(keep_ratio * num_inst)
    # 2. select loss
    top_loss, _ = losses.reshape(-1).topk(num_hns, -1)
    loss_mask = (top_loss != 0)
    # 3. mean loss
    return top_loss[loss_mask].mean()


def focal_loss(y_pred, y_true, gamma: float = 2.0, normalize: bool = False):
    with torch.no_grad():
        p = y_pred.sigmoid()
        pt = (1 - p) * y_true + p * (1 - y_true)
        modulating_factor = pt.pow(gamma)

    if normalize:
        y_pred = y_pred.view(-1)
        y_true = y_true.float().view(-1)
        losses = F.binary_cross_entropy_with_logits(y_pred, y_true, reduction='none')

        modulated_losses = losses * modulating_factor
        scale = losses.sum() / modulated_losses.sum()
        return modulated_losses.sum() * scale
    else:
        return F.binary_cross_entropy_with_logits(y_pred, y_true, modulating_factor, reduction='mean')


@torch.jit.script
def sigmoid_focal_loss(y_pred, y_true, alpha: float = -1, gamma: float = 2, reduction: str = "mean"):
    # implementation of fvcore.nn.loss
    p = torch.sigmoid(y_pred)
    ce_loss = F.binary_cross_entropy_with_logits(
        y_pred, y_true, reduction="none"
    )
    p_t = p * y_true + (1 - p) * (1 - y_true)
    loss = ce_loss * ((1 - p_t) ** gamma)

    if alpha >= 0:
        alpha_t = alpha * y_true + (1 - alpha) * (1 - y_true)
        loss = alpha_t * loss

    if reduction == "mean":
        loss = loss.mean()
    elif reduction == "sum":
        loss = loss.sum()

    return loss


@torch.jit.script
def soft_cross_entropy(input: torch.Tensor, target: torch.Tensor):
    assert input.dim() == 4 and target.dim() == 4
    log_probs = F.log_softmax(input, dim=1)
    return -(target * log_probs).mean(dim=(0, 2, 3)).sum()


class InverseWeightCrossEntroyLoss(nn.Module):
    def __init__(self, class_num, ignore_index=255):
        super(InverseWeightCrossEntroyLoss, self).__init__()
        self.class_num = class_num
        self.ignore_index = ignore_index

    def forward(self, logit, label):
        """
       get inverse cross entropy loss
        Args:
            logit: a tensor, [batch_size, num_class, image_size, image_size]
            label: a tensor, [batch_size, image_size, image_size]
        Returns:
        """
        inverse_weight = self.get_inverse_weight(label)
        cross_entropy = nn.CrossEntropyLoss(weight=inverse_weight, ignore_index=self.ignore_index).cuda()
        inv_w_loss = cross_entropy(logit, label)
        return inv_w_loss

    def get_inverse_weight(self, label):
        mask = (label >= 0) & (label < self.class_num)
        label = label[mask]
        # reduce dim
        total_num = len(label)
        # get unique label, convert unique label to list
        percentage = torch.bincount(label, minlength=self.class_num) / float(total_num)
        # get inverse
        w_for_each_class = 1 / torch.log(1.02 + percentage)
        # convert to tensor
        return w_for_each_class.float()


class StructureLoss(nn.Module):
    """Ref SOD F3Net"""

    def __init__(self, kernel_size=31, stride=1, padding=15):
        super(StructureLoss, self).__init__()
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def _get_weight(self, mask):
        weit = 1 + 5 * torch.abs(F.avg_pool2d(mask, kernel_size=self.kernel_size, stride=self.stride, padding=self.padding) - mask)
        return weit

    def forward(self, pred, mask):
        weit = self._get_weight(mask)

        wbce = F.binary_cross_entropy_with_logits(pred, mask, reduce='none')
        wbce = (weit * wbce).sum(dim=(2, 3)) / weit.sum(dim=(2, 3))

        pred = torch.sigmoid(pred)
        inter = ((pred * mask) * weit).sum(dim=(2, 3))
        union = ((pred + mask) * weit).sum(dim=(2, 3))
        wiou = 1 - (inter + 1) / (union - inter + 1)
        return (wbce + wiou).mean()


class IOULoss(torch.nn.Module):
    def __init__(self, size_average=True):
        super().__init__()
        self.size_average = size_average

    @staticmethod
    def _iou(pred, target, size_average=True):
        b = pred.shape[0]
        IoU = 0.0
        for i in range(0, b):
            # compute the IoU of the foreground
            Iand1 = torch.sum(target[i, :, :, :] * pred[i, :, :, :])
            Ior1 = torch.sum(target[i, :, :, :]) + torch.sum(pred[i, :, :, :]) - Iand1
            IoU1 = Iand1 / Ior1

            # IoU loss is (1-IoU1)
            IoU = IoU + (1 - IoU1)

        return IoU / b

    def forward(self, pred, target):
        return self.iou(pred, target, self.size_average)


class SSIMLoss(nn.Module):
    """# https://github.com/Po-Hsun-Su/pytorch-ssim/blob/master/pytorch_ssim/__init__.py"""
    def __init__(self, window_size=11, size_average=True):
        super().__init__()
        self.window_size = window_size
        self.size_average = size_average
        self.channel = 1
        self.window = self.create_window(window_size, self.channel)

    @staticmethod
    def gaussian(window_size, sigma):
        gauss = torch.Tensor([math.exp(-(x - window_size // 2) ** 2 / float(2 * sigma ** 2)) for x in range(window_size)])
        return gauss / gauss.sum()

    def create_window(self, window_size, channel):
        _1D_window = self.gaussian(window_size, 1.5).unsqueeze(1)
        _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
        window = Variable(_2D_window.expand(channel, 1, window_size, window_size).contiguous())
        return window
    
    @staticmethod
    def ssim(img1, img2, window, window_size, channel, size_average=True):
        mu1 = F.conv2d(img1, window, padding=window_size // 2, groups=channel)
        mu2 = F.conv2d(img2, window, padding=window_size // 2, groups=channel)

        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1 * mu2

        sigma1_sq = F.conv2d(img1 * img1, window, padding=window_size // 2, groups=channel) - mu1_sq
        sigma2_sq = F.conv2d(img2 * img2, window, padding=window_size // 2, groups=channel) - mu2_sq
        sigma12 = F.conv2d(img1 * img2, window, padding=window_size // 2, groups=channel) - mu1_mu2

        C1 = 0.01 ** 2
        C2 = 0.03 ** 2

        ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))

        if size_average:
            return ssim_map.mean()
        else:
            return ssim_map.mean(1).mean(1).mean(1)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        if channel == self.channel and self.window.data.type() == img1.data.type():
            window = self.window
        else:
            window = self.create_window(self.window_size, channel)

            if img1.is_cuda:
                window = window.cuda(img1.get_device())
            window = window.type_as(img1)

            self.window = window
            self.channel = channel

        return self.ssim(img1, img2, window, self.window_size, channel, self.size_average)


class BCE_SSIM_IoU_Loss(nn.Module):
    """Ref BASNet"""
    def __init__(self, window_size=11):
        super(BCE_SSIM_IoU_Loss, self).__init__()

        self.bce = nn.BCEWithLogitsLoss()
        self.ssim = SSIMLoss(window_size=window_size)
        self.iou = IOULoss()

    def forward(self, pred, target):
        bce_out = self.bce(pred, target)
        ssim_out = 1 - self.ssim(pred, target)
        iou_out = self.iou(pred, target)
        loss = bce_out + ssim_out + iou_out

        return loss



# -*- coding: utf-8 -*-
"""
Copyright (c) Yuwei Jin
Created on 2020-11-13 16:18
Written by Yuwei Jin (642281525@qq.com)
"""
import math
import bisect
import weakref
from functools import wraps
from typing import Union, Any, Optional

from torch.optim import Optimizer


def _map_value_to_list(input: Optional[Union[float, tuple, list]], target_list: list) -> list:
    """
    map the given value to a list that has the same length of target list.
    Note, the mapped elements in value would become float type.
    """
    if isinstance(input, tuple):  # covert tuple to list
        input = list(input)
    # when applying different gamma values for different param groups, the dimention between gamma and the base lrs of optimizer must be same!
    if isinstance(input, list):
        assert len(input) == len(target_list)
        value = list(map(float, input))
    else:
        input = float(input)
        value = [input for _ in target_list]

    return value


class _LRScheduler:
    def __init__(self,
                 optimizer: Optimizer,
                 adjust_period: Optional[int] = None,
                 eta_min_lr: Optional[Union[float, tuple, list]] = None
                 ):
        """Base class for learning scheduler class.
            Args:
                optimizer: Wrapped optimizer.
                adjust_period: Maximum number of adjust learning rate. Default: None.
                eta_min_lr: Minimum learning rate. Default: 0.
        """

        # Attach optimizer
        if not isinstance(optimizer, Optimizer):
            raise TypeError('{} is not an Optimizer'.format(type(optimizer).__name__))
        self.optimizer = optimizer

        self.adjust_period = int(adjust_period) if adjust_period is not None else 0
        self.warm_base_lrs = []
        self.warmup_period = 0  # warmup epochs

        # set initial learning rate for optimizer
        self.last_epoch = 1  # 从last_start开始后已经记录了多少个epoch
        if self.last_epoch == 1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified in param_groups[{}] when resuming an optimizer".format(i))

        # Following https://github.com/pytorch/pytorch/issues/20124
        # We would like to ensure that `train.step()` is called after
        # `optimizer.step()`
        def with_counter(method):
            if getattr(method, '_with_counter', False):
                # `optimizer.step()` has already been replaced, return.
                return method

            # Keep a weak reference to the optimizer instance to prevent
            # cyclic references.
            instance_ref = weakref.ref(method.__self__)
            # Get the unbound method for the same purpose.
            func = method.__func__
            cls = instance_ref().__class__
            del method

            @wraps(func)
            def wrapper(*args, **kwargs):
                instance = instance_ref()
                instance._step_count += 1
                wrapped = func.__get__(instance, cls)
                return wrapped(*args, **kwargs)

            # Note that the returned core here is no longer a bound method,
            # so attributes like `__func__` and `__self__` no longer exist.
            wrapper._with_counter = True
            return wrapper

        self.optimizer.step = with_counter(self.optimizer.step)
        self.optimizer._step_count = 0

        # get the base lrs of optimizer
        self.base_lrs = list(map(lambda group: group['initial_lr'], self.optimizer.param_groups))

        # set eta_min_lrs
        eta_min_lr = eta_min_lr if eta_min_lr is not None else 0.0
        eta_min_lrs = _map_value_to_list(eta_min_lr, self.base_lrs)
        for eta_min_lr, base_lr in zip(eta_min_lrs, self.base_lrs):
            assert eta_min_lr < base_lr
        self.eta_min_lrs = eta_min_lrs

    def warmup_init(self, warm_base_lr: Union[float, list, tuple], warmup_period: int):
        """Set the begin learning rate and warmup period for warm up.
            Args:
                warm_base_lr: Beginning learning rate of warm_up
                warmup_period: Number of warm_up epochs.
        """
        assert warmup_period >= 0
        assert warm_base_lr is not None

        # ensure warmup_begin_lr is list or float type.
        if warmup_period > 0:
            warmup_base_lrs = _map_value_to_list(warm_base_lr, self.base_lrs)
            for warm_base_lr, base_lr in zip(warmup_base_lrs, self.base_lrs):
                assert warm_base_lr <= base_lr

            self.warm_base_lrs = warmup_base_lrs
            self.warmup_period = warmup_period

            for param_group, lr in zip(self.optimizer.param_groups, warmup_base_lrs):
                param_group['lr'] = lr

    def get_lr(self) -> list or tuple:
        # Compute learning rate using chainable form of the scheduler
        raise NotImplemented

    def get_warmup_lr(self) -> list:
        """Return the computed learning rate by current scheduler"""
        lrs = []
        for warm_base_lr, base_lr in zip(self.warm_base_lrs, self.base_lrs):
            lr = warm_base_lr + self.last_epoch * (base_lr - warm_base_lr) / self.warmup_period
            lrs.append(lr)
        return lrs

    def update_lr(self) -> None:
        values = self.get_lr()
        for data in zip(self.optimizer.param_groups, values):
            param_group, lr = data
            param_group['lr'] = lr

    def get_last_lr(self):
        return list(map(lambda group: group['lr'], self.optimizer.param_groups))

    def step(self, epoch=None) -> None:
        """Update the learning rate of next epoch. Note: it mast be called after optimizer.step().
            Args:
                epoch: Current epoch. Default: None.
        """
        # if self.optimizer._step_count == 0:
        #     warning ("Detected call of `train.step()` before `optimizer.step()`. ")

        if epoch is None:
            self.last_epoch += 1
            self.update_lr()
        else:
            if epoch < self.last_epoch:
                raise ValueError(
                    'Input epoch should be less than last.epoch. Get epoch: {}, last_epoch: {}'.format(epoch,
                                                                                                       self.last_epoch))
            self.last_epoch = epoch
            self.update_lr()


class ExponentialLR(_LRScheduler):
    """Decays the learning rate of each parameter group by gamma every epoch"""

    def __init__(self,
                 optimizer: Optimizer,
                 gamma: Optional[Union[float, list, tuple]] = 0.9,
                 eta_min_lr: Optional[Union[float, list, tuple]] = None,
                 adjust_period: int = None,
                 warm_period: int = 0,
                 warm_start_lr: Optional[Union[float, list, tuple]] = 1e-6
                 ):
        """
        Args:
            optimizer: Wrapped optimizer.
            gamma: Multiplicative factor of learning rate decay.
            eta_min_lr: learning rate. When the original learning rate is decayed to eta_min, the decay process is stopped.
            adjust_period: Maximum number of update epochs.
        """
        super(ExponentialLR, self).__init__(optimizer=optimizer, eta_min_lr=eta_min_lr, adjust_period=adjust_period)

        # when applying different gamma values for different param groups,
        # the dimension between gamma and the base lrs of optimizer must be same!

        self.warmup_init(warm_start_lr, warm_period)
        self.gamma = _map_value_to_list(gamma, self.base_lrs)

    def _get_closed_form_lr(self, curr_epoch) -> list:
        lrs = [(base_lr - eta_min_lr) * pow(gamma, curr_epoch) + eta_min_lr for gamma, base_lr, eta_min_lr in
               zip(self.gamma, self.base_lrs, self.eta_min_lrs)]
        return lrs

    def get_lr(self):
        if self.last_epoch <= self.warmup_period:
            return self.get_warmup_lr()
        else:
            if self.adjust_period == 0:
                curr_epoch = self.last_epoch - self.warmup_period
                return self._get_closed_form_lr(curr_epoch)
            else:
                if self.warmup_period < self.last_epoch <= self.adjust_period:
                    curr_epoch = self.last_epoch - self.warmup_period
                    return self._get_closed_form_lr(curr_epoch)
                else:
                    return self.get_last_lr()


class PolyLR(_LRScheduler):
    """Decays the learning rate of each parameter group by polynomial strategy every epoch"""

    def __init__(self,
                 optimizer: Optimizer,
                 max_iterations: int,
                 power: Optional[Union[float, list, tuple]] = 0.9,
                 eta_min_lr: Optional[Union[float, list, tuple]] = None,
                 warm_period: int = 0,
                 warm_start_lr: Optional[Union[float, list, tuple]] = 1e-6
                 ):
        """
            Args:
                optimizer: Wrapped optimizer.
                power: Multiplicative factor of learning rate decay.
                eta_min_lr: Minimum learning rate. When the original learning rate is decayed to eta_min, the decay
                           process is stopped.
                max_iterations number of update epochs.
        """
        assert max_iterations > 0
        super(PolyLR, self).__init__(optimizer, eta_min_lr=eta_min_lr, adjust_period=max_iterations)

        self.power = _map_value_to_list(power, self.base_lrs)
        self.warmup_init(warm_start_lr, warm_period)

    def _get_closed_form_lr(self, curr_epoch) -> list:
        lrs = []
        for power, base_lr, eta_min_lr in zip(self.power, self.base_lrs, self.eta_min_lrs):
            lr = (base_lr - eta_min_lr) * (1 - float(curr_epoch) / self.adjust_period) ** power + eta_min_lr
            lrs.append(lr)
        return lrs

    def get_lr(self):
        if self.last_epoch <= self.warmup_period:
            return self.get_warmup_lr()
        elif self.warmup_period < self.last_epoch <= self.adjust_period:
            curr_epoch = self.last_epoch - self.warmup_period
            return self._get_closed_form_lr(curr_epoch)
        else:
            return self.get_last_lr()


class StepLR(_LRScheduler):
    """Decays the learning rate of each parameter group by gamma every step_size epochs."""

    def __init__(self,
                 optimizer: Optimizer,
                 step_size: int,
                 gamma: float = 0.1,
                 warm_period: int = 0,
                 warm_start_lr: Optional[Union[float, list, tuple]] = 1e-6
                 ):
        """
            Args:
                optimizer: Wrapped optimizer.
                step_size: Period of learning rate decay.
                gamma: Multiplicative factor of learning rate decay.
        """
        super(StepLR, self).__init__(optimizer, adjust_period=None, eta_min_lr=None)
        self.step_size = step_size
        self.gamma = float(gamma)
        self.warmup_init(warm_base_lr=warm_start_lr, warmup_period=warm_period)

    def get_lr(self):
        if self.last_epoch <= self.warmup_period:
            return self.get_warmup_lr()
        else:
            curr_epoch = self.last_epoch - self.warmup_period
            lrs = [base_lr * self.gamma ** (curr_epoch // self.step_size) for base_lr in self.base_lrs]
            return lrs


class MultiStepLR(_LRScheduler):
    def __init__(self,
                 optimizer: Optimizer,
                 milestones: Union[tuple, list],
                 gamma: float = 0.1,
                 warm_period: int = 0,
                 warm_start_lr: Optional[Union[float, list, tuple]] = 1e-6
                 ):
        """Decays the learning rate of each parameter group by gamma once the number of epoch reaches one of the
        milestones.
            Args:
                milestones: List of epoch indices. Must be increasing.
                gamma:  Multiplicative factor of learning rate decay. Default: 0.1.
        """
        super(MultiStepLR, self).__init__(optimizer=optimizer, adjust_period=None, eta_min_lr=None)

        self.milestones = list(sorted(milestones))
        self.gamma = float(gamma)

        self.warmup_init(warm_base_lr=warm_start_lr, warmup_period=warm_period)

    def get_lr(self):
        if self.last_epoch <= self.warmup_period:
            return self.get_warmup_lr()
        else:
            curr_epoch = self.last_epoch - self.warmup_period
            lrs = [base_lr * self.gamma ** bisect.bisect_right(self.milestones, curr_epoch) for base_lr in
                   self.base_lrs]
            return lrs


class CosineLR(_LRScheduler):
    """Set the learning rate of each parameter group using a cosine annealing schedule"""

    def __init__(self,
                 optimizer,
                 T_max: int,
                 eta_min_lr: Optional[Union[list, tuple, float]] = None,
                 warm_period: int = 5,
                 warm_start_lr: Optional[Union[float, list, tuple]] = 1e-6
                 ):
        """
        Args:
            optimizer (Optimizer): Wrapped optimizer.
            eta_min_lr (float): Minimum learning rate.
            T_max: Period of learning rate decay.
        """
        super(CosineLR, self).__init__(optimizer=optimizer, adjust_period=T_max, eta_min_lr=eta_min_lr)
        self.T_max = T_max
        self.warmup_init(warm_base_lr=warm_start_lr, warmup_period=warm_period)

    def get_lr(self):
        if self.last_epoch <= self.warmup_period:
            return self.get_warmup_lr()
        elif self.warmup_period < self.last_epoch <= self.adjust_period:
            # curr_lr = self.optimizer.state_dict()['param_groups'][0]['lr_scheduler']
            curr_epoch = self.last_epoch - self.warmup_period
            T_max = self.T_max - self.warmup_period
            lrs = []
            for base_lr, eta_min in zip(self.base_lrs, self.eta_min_lrs):
                lr = eta_min + 0.5 * (base_lr - eta_min) * (1 + math.cos(curr_epoch * math.pi / T_max))
                lrs.append(lr)
            return lrs
        else:
            return self.get_last_lr()


class STANetLR(_LRScheduler):
    def __init__(self, total_epoch: int, start_epoch: int, optimizer: Optimizer):
        """
        :param start_epoch: iter at starting learning rate
        """
        super().__init__(optimizer)

        self.total_epoch = total_epoch
        self.start_epoch = start_epoch

    def get_lr(self):
        if self.last_epoch < self.start_epoch:
            return self.base_lrs
        else:
            lrs = []
            for base_lr in self.base_lrs:
                d = self.last_epoch - self.start_epoch
                lr = base_lr - d * base_lr / (self.total_epoch - self.start_epoch)
                lrs.append(lr)
            return lrs


if __name__ == '__main__':
    import torch.nn as nn


    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()

            self.a = nn.Conv2d(3, 3, 1)
            self.b = nn.Conv2d(3, 3, 1)

        def forward(self, x):
            x = self.a(x)
            x = self.b(x)
            return x


    import torch

    model = Net()

    epochs = 60
    steps = 800
    init_lr = 0.001

    lr = []
    x = []

    params = [
        {'params': model.a.parameters(), 'lr': init_lr},
        {'params': model.b.parameters(), 'lr': 10 * init_lr}
    ]
    optimizer = torch.optim.Adam(params=params, weight_decay=0.0005)

    lr_scheduler = STANetLR(optimizer=optimizer, start_epoch=20, total_epoch=epochs)

    k = 0

    for i in range(epochs):
        for j in range(steps):
            optimizer.zero_grad()
            optimizer.step()

        curr_lr = optimizer.state_dict()['param_groups'][0]['lr']
        lr.append(curr_lr)
        k += 1
        x.append(k)

        print("the {} iter: {:.10f}".format(k, curr_lr))

        lr_scheduler.step()

    from matplotlib import pyplot as plt

    plt.plot(x, lr, 'r', linewidth=2)
    plt.ylabel('learning rate')
    plt.xlabel('epochs')
    plt.grid()
    plt.show()



# -*- coding: UTF-8 -*-
"""
Copyright (c) Yuwei Jin
Created at 2021-10-23 18:24 
Written by Yuwei Jin (642281525@qq.com)
"""
import torch
from torch import nn
from torch.nn import functional as F

from sv.libs.tools.init_weight import init_weights
from sv.modules.packaged_conv import Conv1x1_BN_ReLU, Conv3x3_BN_ReLU, Conv3x3, DWConv, ConvBNReLU, Conv1x1


class DoubleConv(nn.Module):
    def __init__(self, in_chs, out_chs):
        super(DoubleConv, self).__init__()
        self.conv = nn.Sequential(
            Conv3x3_BN_ReLU(in_chs, out_chs),
            Conv3x3_BN_ReLU(out_chs, out_chs)
        )

    def forward(self, x):
        return self.conv(x)


class UNetEncBlock(nn.Module):
    """ref U-Net"""

    def __init__(self, in_chs, out_chs):
        super().__init__()
        self.conv = DoubleConv(in_chs, out_chs)
        self.maxpool = nn.MaxPool2d(2)

    def forward(self, x):
        x1 = self.conv(x)
        x2 = self.maxpool(x1)
        return x1, x2


class UNetDecBlock(nn.Module):
    """original U-Net decoder block"""

    def __init__(self, in_chs_l, in_chs_h):
        super().__init__()

        if in_chs_l != in_chs_h:
            self.reduction_conv = Conv1x1_BN_ReLU(in_chs_h, in_chs_l)

        self.conv = DoubleConv(in_chs_l*2, in_chs_l)

    def forward(self, low_level_f, high_level_f):
        if hasattr(self, 'reduction_conv'):
            high_level_f = self.reduction_conv(high_level_f)

        if high_level_f.size()[2:] != low_level_f.size()[2:]:
            high_level_f = F.interpolate(high_level_f, size=low_level_f.size()[2:], mode='bilinear', align_corners=True)

        x = torch.cat([low_level_f, high_level_f], dim=1)

        out = self.conv(x)

        return out


# ----- for classification -----
class Classifier(nn.Module):
    """
    Convolution classify module.
    """

    def __init__(self, in_channels, num_classes):
        super(Classifier, self).__init__()

        self.conv = nn.Conv2d(in_channels, num_classes, kernel_size=1)

        init_weights(self.conv)

    def forward(self, x):
        return self.conv(x)


class Aux_Module(nn.Module):
    """
    auxiliary criterion module for resnet deep supervision.
    """

    def __init__(self, in_channels, mid_channels, num_classes=1, drop_out=0.1):
        super(Aux_Module, self).__init__()

        self.aux = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Dropout2d(p=drop_out),
            nn.Conv2d(mid_channels, num_classes, kernel_size=1)
        )
        init_weights(self.aux)

    def forward(self, x):
        res = self.aux(x)
        return res


# --------- basic residual module
class BasicResidualBlock(nn.Module):
    def __init__(self, in_channel, out_channel, stride=1, downsample=None):
        super().__init__()

        self.conv1 = nn.Conv2d(in_channel, out_channel, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(out_channel)
        self.conv2 = nn.Conv2d(out_channel, out_channel, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_channel)
        self.relu = nn.ReLU(inplace=True)

        if downsample is not None:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_channel, out_channel, 1, bias=False),
                nn.BatchNorm2d(out_channel)
            )
            init_weights(self.downsample)

        init_weights(self.conv1, self.bn1, self.conv2, self.bn2)

    def forward(self, x):
        identity = x
        if hasattr(self, 'downsample'):
            identity = self.downsample(x)

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        if downsample is not None:
            self.downsample = nn.Sequential(
                nn.Conv2d(inplanes, planes * 4, 1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * 4)
            )
            init_weights(self.downsample)

        init_weights(self.conv1, self.bn1, self.conv2, self.bn2, self.conv3, self.bn3)
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if hasattr(self, 'downsample'):
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class AlignModule(nn.Module):
    """
    Align module for feature align. Borrowed from SFNet.
    """

    def __init__(self, inplane, outplane, kernel_size=3):
        super(AlignModule, self).__init__()
        self.down_h = nn.Conv2d(inplane, outplane, kernel_size, bias=False)
        self.down_l = nn.Conv2d(inplane, outplane, kernel_size, bias=False)
        self.flow_make = nn.Conv2d(outplane * 2, 2, kernel_size=3, padding=1, bias=False)

        init_weights(self.down_h, self.down_l, self.flow_make)

    def forward(self, low_feature, h_feature):
        h_feature_orign = h_feature
        h, w = low_feature.size()[2:]
        size = (h, w)
        low_feature = self.down_l(low_feature)
        h_feature = self.down_h(h_feature)
        h_feature = F.interpolate(h_feature, size=size, mode="bilinear", align_corners=False)
        flow = self.flow_make(torch.cat([h_feature, low_feature], 1))
        h_feature = self.flow_warp(h_feature_orign, flow, size=size)

        return h_feature

    @staticmethod
    def flow_warp(input, flow, size):
        out_h, out_w = size
        n, c, h, w = input.size()

        norm = torch.tensor([[[[out_w, out_h]]]]).type_as(input).to(input.device)

        w = torch.linspace(-1.0, 1.0, out_h).view(-1, 1).repeat(1, out_w)

        h = torch.linspace(-1.0, 1.0, out_w).repeat(out_h, 1)

        grid = torch.cat((h.unsqueeze(2), w.unsqueeze(2)), 2)
        grid = grid.repeat(n, 1, 1, 1).type_as(input).to(input.device)
        grid = grid + flow.permute(0, 2, 3, 1) / norm

        output = F.grid_sample(input, grid, align_corners=True)
        return output


class RefinementResidualBlock(nn.Module):
    """ ref: Discriminative Feature Network """

    def __init__(self, in_chs, out_chs):
        super(RefinementResidualBlock, self).__init__()
        self.conv1 = Conv1x1_BN_ReLU(in_chs, out_chs)

        if in_chs > 256:
            self.dropout = nn.Dropout2d(0.1)

        self.conv2 = nn.Sequential(
            Conv3x3_BN_ReLU(out_chs, out_chs),
            Conv3x3(out_chs, out_chs),
            nn.BatchNorm2d(out_chs)
        )

        self.identity = nn.Identity()

        init_weights(self.conv1, self.conv2)

    def forward(self, x):
        x = self.conv1(x)
        if hasattr(self, 'dropout'):
            x = self.dropout(x)
        identity = self.identity(x)
        out = self.conv2(x)
        out += identity
        out = torch.relu(out)
        return out


class CAB(nn.Module):
    ''' Channel Attention Block
    Learning a Discriminative Feature Network for Semantic Segmentation (CVPR2018)(face++)
    '''

    def __init__(self, in_channels, out_channels):
        super(CAB, self).__init__()
        self.global_pooling = nn.AdaptiveAvgPool2d(1)
        self.conv1 = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0)
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv2d(out_channels, out_channels, kernel_size=1, stride=1, padding=0)
        self.sigmod = nn.Sigmoid()

    def forward(self, x):
        x1, x2 = x  # high, low
        x = torch.cat([x1, x2], dim=1)
        x = self.global_pooling(x)
        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.sigmod(x)
        x2 = x * x2
        res = x2 + x1
        return res


class AttentionRefinementModule(nn.Module):
    """ref bisenet_v1"""

    def __init__(self, in_chan, out_chan, *args, **kwargs):
        super(AttentionRefinementModule, self).__init__()
        self.conv = Conv3x3_BN_ReLU(in_chan, out_chan)
        self.conv_atten = nn.Conv2d(out_chan, 1, kernel_size=1, bias=False)
        self.bn_atten = nn.BatchNorm2d(out_chan)
        self.sigmoid_atten = nn.Sigmoid()

        init_weights(self.conv_atten, self.bn_atten)

    def forward(self, x):
        feat = self.conv(x)
        atten = self.conv_atten(feat)
        atten = self.bn_atten(atten)
        atten = self.sigmoid_atten(atten)
        out = torch.mul(feat, atten)
        return out


class CEBlock(nn.Module):
    """
    Context embedding block
    Ref: BiSeNet_v2
    """

    def __init__(self, in_chs):
        super(CEBlock, self).__init__()

        self.gap = nn.AdaptiveAvgPool2d(1)
        self.conv_gap = ConvBNReLU(in_chs, in_chs, 1, stride=1, padding=0)
        self.arm = AttentionRefinementModule(in_chs, in_chs)

        # TODO: in paper here is naive conv2d, no bn-relu
        self.conv_last = DWConv(in_chs, in_chs, 3)

    def forward(self, x):
        feat = self.gap(x)
        feat = self.conv_gap(feat)
        feat = F.interpolate(feat, x.size()[2:], mode='bilinear', align_corners=True)
        feat = feat + self.arm(feat)
        feat = self.conv_last(feat)
        return feat


class SelectiveRefineModule(nn.Module):
    """ref BiseNetv1"""

    def __init__(self, in_chs_l, in_chs_h, out_chs):
        super(SelectiveRefineModule, self).__init__()

        self.fusion_conv = Conv1x1_BN_ReLU(in_chs_l + in_chs_h, out_chs)

        self.conv1 = Conv3x3(out_chs, out_chs, dilation=1)
        self.conv2 = Conv3x3(out_chs, out_chs, dilation=2)

        self.att = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            Conv1x1(out_chs, out_chs, bias=False),
            nn.Softmax(dim=1)
        )

    def forward(self, low_f, high_f):
        x = torch.cat([low_f, high_f], dim=1)
        x = self.fusion_conv(x)

        x1 = self.conv1(x)
        x2 = self.conv2(x)

        y = x1 + x2

        att = self.att(y)

        out = att * x1 + att * x2

        return out


class FSModule(nn.Module):
    """
    Ref:  Foreground-Aware Relation Network for Geospatial Object Segmentation in High Spatial Resolution Remote Sensing Imagery
    """

    def __init__(self, in_chs, out_chs):
        super(FSModule, self).__init__()
        self.conv1 = Conv1x1_BN_ReLU(in_chs, out_chs)
        self.conv2 = Conv1x1_BN_ReLU(in_chs, out_chs)

    def forward(self, v, u):
        x = self.conv1(v)
        r = torch.mul(x, u)
        k = self.conv2(v)
        z = k / (1 + torch.exp(-r))
        return z


class GlobalAttentionUpsample(nn.Module):
    """ref: Pyramid Attention Network for Semantic Segmentation"""

    def __init__(self, channels_high, channels_low, upsample=True):
        super(GlobalAttentionUpsample, self).__init__()

        self.upsample = upsample

        self.conv3x3 = nn.Sequential(
            nn.Conv2d(channels_low, channels_low, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(channels_low)
        )

        self.conv1x1 = nn.Sequential(
            nn.Conv2d(channels_high, channels_low, kernel_size=1, padding=0, bias=False),
            nn.BatchNorm2d(channels_low)
        )

        if upsample:
            self.conv_upsample = nn.Sequential(
                nn.ConvTranspose2d(channels_high, channels_low, kernel_size=4, stride=2, padding=1, bias=False),
                nn.BatchNorm2d(channels_low)
            )
        else:
            self.conv_reduction = nn.Sequential(
                nn.Conv2d(channels_high, channels_low, kernel_size=1, padding=0, bias=False),
                nn.BatchNorm2d(channels_low)
            )

        self.relu = nn.ReLU(inplace=True)

        init_weights(self.conv_reduction, self.conv_upsample, self.conv1x1, self.conv3x3)

    def forward(self, fms_high, fms_low, fm_mask=None):
        """
        Use the high level features with abundant catagory information to weight the low level features with pixel
        localization information. In the meantime, we further use mask feature maps with catagory-specific information
        to localize the mask position.
        :param fms_high: Features of high level. Tensor.
        :param fms_low: Features of low level.  Tensor.
        :param fm_mask:
        :return: fms_att_upsample
        """
        b, c, h, w = fms_high.shape

        fms_high_gp = nn.AvgPool2d(fms_high.shape[2:])(fms_high).view(len(fms_high), c, 1, 1)
        fms_high_gp = self.conv1x1(fms_high_gp)
        fms_high_gp = self.relu(fms_high_gp)

        # fms_low_mask = torch.cat([fms_low, fm_mask], dim=1)
        fms_low_mask = self.conv3x3(fms_low)

        fms_att = fms_low_mask * fms_high_gp
        if self.upsample:
            out = self.relu(self.conv_upsample(fms_high)) + fms_att
        else:
            out = self.relu(self.conv_reduction(fms_high)) + fms_att

        return out


def upsample(x, size, mode='bilinear'):
    if mode == 'bilinear':
        out = F.interpolate(x, size=size, mode=mode, align_corners=True)
    elif mode == 'nearest':
        out = F.interpolate(x, size=size, mode=mode)
    else:
        raise NotImplementedError('Upsample mode is not supported! Got mode: {}'.format(mode))

    return out